@@ -12,6 +12,7 @@ DEFAULT_OPTIONS = {
         'instance': 'default',
         'dry_run': False,
         'pid_file_dir': '/tmp',
+        'run_once': False,
 
         # [global]
         'region': 'us-east-1',
@@ -48,6 +49,10 @@ DEFAULT_OPTIONS = {
         'max_provisioned_reads': None,
         'min_provisioned_writes': None,
         'max_provisioned_writes': None,
+        'num_read_checks_before_scale_down': 1,
+        'num_write_checks_before_scale_down': 1,
+        'num_read_checks_reset_percent': 0,
+        'num_write_checks_reset_percent': 0,
         'allow_scaling_down_reads_on_0_percent': False,
         'allow_scaling_down_writes_on_0_percent': False,
         'always_decrease_rw_together': False,
@@ -76,6 +81,10 @@ DEFAULT_OPTIONS = {
         'max_provisioned_reads': None,
         'min_provisioned_writes': None,
         'max_provisioned_writes': None,
+        'num_read_checks_before_scale_down': 1,
+        'num_write_checks_before_scale_down': 1,
+        'num_read_checks_reset_percent': 0,
+        'num_write_checks_reset_percent': 0,
         'allow_scaling_down_reads_on_0_percent': False,
         'allow_scaling_down_writes_on_0_percent': False,
         'always_decrease_rw_together': False,
@@ -341,7 +350,7 @@ def __check_gsi_rules(configuration):
                 'min_provisioned_reads',
                 'max_provisioned_reads',
                 'min_provisioned_writes',
-                'max_provisioned_writes'
+                'max_provisioned_writes',
             ]
             for option in options:
                 if gsi[option] < 1:
@@ -430,7 +439,7 @@ def __check_table_rules(configuration):
                 'than 100% at a time. Setting --increase-reads-with to 100.')
             table['increase_reads_with'] = 100
 
-        # Ensure values > 1 for some important configuration options
+        # Ensure values > 0 for some important configuration options
         options = [
             'reads_lower_threshold',
             'reads_upper_threshold',
@@ -443,7 +452,9 @@ def __check_table_rules(configuration):
             'min_provisioned_reads',
             'max_provisioned_reads',
             'min_provisioned_writes',
-            'max_provisioned_writes'
+            'max_provisioned_writes',
+            'num_read_checks_before_scale_down',
+            'num_write_checks_before_scale_down'
         ]
         for option in options:
             if table[option] < 1:
