@@ -1,17 +1,18 @@
+# -*- coding: utf-8 -*-
 """ Core components """
-import datetime
-
 from boto.exception import JSONResponseError, BotoServerError
 
-from dynamic_dynamodb.calculators import gsi as calculators
+from dynamic_dynamodb import calculators
+from dynamic_dynamodb.aws import dynamodb
 from dynamic_dynamodb.core import circuit_breaker
-from dynamic_dynamodb.core import dynamodb
 from dynamic_dynamodb.statistics import gsi as gsi_stats
 from dynamic_dynamodb.log_handler import LOGGER as logger
 from dynamic_dynamodb.config_handler import get_global_option, get_gsi_option
 
 
-def ensure_provisioning(table_name, table_key, gsi_name, gsi_key):
+def ensure_provisioning(
+        table_name, table_key, gsi_name, gsi_key,
+        num_consec_read_checks, num_consec_write_checks):
     """ Ensure that provisioning is correct for Global Secondary Indexes
 
     :type table_name: str
@@ -22,27 +23,42 @@ def ensure_provisioning(table_name, table_key, gsi_name, gsi_key):
     :param gsi_name: Name of the GSI
     :type gsi_key: str
     :param gsi_key: GSI configuration option key name
+    :type num_consec_read_checks: int
+    :param num_consec_read_checks: How many consecutive checks have we had
+    :type num_consec_write_checks: int
+    :param num_consec_write_checks: How many consecutive checks have we had
+    :returns: (int, int) -- num_consec_read_checks, num_consec_write_checks
     """
     if get_global_option('circuit_breaker_url'):
         if circuit_breaker.is_open():
             logger.warning('Circuit breaker is OPEN!')
-            return None
+            return (0, 0)
 
     logger.info(
         '{0} - Will ensure provisioning for global secondary index {1}'.format(
             table_name, gsi_name))
 
     try:
-        read_update_needed, updated_read_units = __ensure_provisioning_reads(
-            table_name,
-            table_key,
-            gsi_name,
-            gsi_key)
-        write_update_needed, updated_write_units = __ensure_provisioning_writes(
-            table_name,
-            table_key,
-            gsi_name,
-            gsi_key)
+        read_update_needed, updated_read_units, num_consec_read_checks = \
+            __ensure_provisioning_reads(
+                table_name,
+                table_key,
+                gsi_name,
+                gsi_key,
+                num_consec_read_checks)
+        write_update_needed, updated_write_units, num_consec_write_checks = \
+            __ensure_provisioning_writes(
+                table_name,
+                table_key,
+                gsi_name,
+                gsi_key,
+                num_consec_write_checks)
+
+        if read_update_needed:
+            num_consec_read_checks = 0
+
+        if write_update_needed:
+            num_consec_write_checks = 0
 
         # Handle throughput updates
         if read_update_needed or write_update_needed:
@@ -67,6 +83,10 @@ def ensure_provisioning(table_name, table_key, gsi_name, gsi_key):
                     gsi_name))
     except JSONResponseError:
         raise
+    except BotoServerError:
+        raise
+
+    return num_consec_read_checks, num_consec_write_checks
 
 
 def __calculate_always_decrease_rw_values(
@@ -114,7 +134,8 @@ def __calculate_always_decrease_rw_values(
     return (read_units, write_units)
 
 
-def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
+def __ensure_provisioning_reads(
+        table_name, table_key, gsi_name, gsi_key, num_consec_read_checks):
     """ Ensure that provisioning is correct
 
     :type table_name: str
@@ -125,7 +146,10 @@ def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
     :param gsi_name: Name of the GSI
     :type gsi_key: str
     :param gsi_key: Configuration option key name
-    :returns: (bool, int) -- update_needed, updated_read_units
+    :type num_consec_read_checks: int
+    :param num_consec_read_checks: How many consecutive checks have we had
+    :returns: (bool, int, int)
+        update_needed, updated_read_units, num_consec_read_checks
     """
     if not get_gsi_option(table_key, gsi_key, 'enable_reads_autoscaling'):
         logger.info(
@@ -133,11 +157,11 @@ def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
             'Autoscaling of reads has been disabled'.format(
                 table_name, gsi_name))
         return False, dynamodb.get_provisioned_gsi_read_units(
-            table_name, gsi_name)
+            table_name, gsi_name), 0
 
     update_needed = False
     try:
-        updated_read_units = dynamodb.get_provisioned_gsi_read_units(
+        current_read_units = dynamodb.get_provisioned_gsi_read_units(
             table_name, gsi_name)
         consumed_read_units_percent = \
             gsi_stats.get_consumed_read_units_percent(table_name, gsi_name)
@@ -160,11 +184,34 @@ def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
                 table_key, gsi_key, 'throttled_reads_upper_threshold')
         max_provisioned_reads = \
             get_gsi_option(table_key, gsi_key, 'max_provisioned_reads')
+        num_read_checks_before_scale_down = \
+            get_gsi_option(
+                table_key, gsi_key, 'num_read_checks_before_scale_down')
+        num_read_checks_reset_percent = \
+            get_gsi_option(table_key, gsi_key, 'num_read_checks_reset_percent')
     except JSONResponseError:
         raise
     except BotoServerError:
         raise
 
+    # Set the updated units to the current read unit value
+    updated_read_units = current_read_units
+
+    # Reset consecutive reads if num_read_checks_reset_percent is reached
+    if num_read_checks_reset_percent:
+
+        if consumed_read_units_percent >= num_read_checks_reset_percent:
+
+            logger.info(
+                '{0} - Resetting the number of consecutive '
+                'read checks. Reason: Consumed percent {1} is '
+                'greater than reset percent: {2}'.format(
+                    table_name,
+                    consumed_read_units_percent,
+                    num_read_checks_reset_percent))
+
+            num_consec_read_checks = 0
+
     if (consumed_read_units_percent == 0 and not
             get_gsi_option(
                 table_key,
@@ -177,74 +224,78 @@ def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
                 table_name, gsi_name))
 
     elif consumed_read_units_percent >= reads_upper_threshold:
-        if increase_reads_unit == 'percent':
-            updated_provisioning = calculators.increase_reads_in_percent(
-                updated_read_units,
-                increase_reads_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key,)
-        else:
-            updated_provisioning = calculators.increase_reads_in_units(
-                updated_read_units,
-                increase_reads_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
 
-        if updated_read_units != updated_provisioning:
+        if increase_reads_unit == 'percent':
+            calculated_provisioning = calculators.increase_reads_in_percent(
+                current_read_units,
+                increase_reads_with,
+                get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
+        else:
+            calculated_provisioning = calculators.increase_reads_in_units(
+                current_read_units,
+                increase_reads_with,
+                get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
+
+        if current_read_units != calculated_provisioning:
+            logger.info(
+                '{0} - Resetting the number of consecutive '
+                'read checks. Reason: scale up event detected'.format(
+                    table_name))
+            num_consec_read_checks = 0
             update_needed = True
-            updated_read_units = updated_provisioning
+            updated_read_units = calculated_provisioning
 
     elif throttled_read_count > throttled_reads_upper_threshold:
 
         if throttled_reads_upper_threshold > 0:
 
             if increase_reads_unit == 'percent':
-                updated_provisioning = calculators.increase_reads_in_percent(
-                    updated_read_units,
+                calculated_provisioning = calculators.increase_reads_in_percent(
+                    current_read_units,
                     increase_reads_with,
-                    table_name,
-                    table_key,
-                    gsi_name,
-                    gsi_key,)
+                    get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                    '{0} - GSI: {1}'.format(table_name, gsi_name))
             else:
-                updated_provisioning = calculators.increase_reads_in_units(
-                    updated_read_units,
+                calculated_provisioning = calculators.increase_reads_in_units(
+                    current_read_units,
                     increase_reads_with,
-                    table_name,
-                    table_key,
-                    gsi_name,
-                    gsi_key)
+                    get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                    '{0} - GSI: {1}'.format(table_name, gsi_name))
 
-            if updated_read_units != updated_provisioning:
+            if current_read_units != calculated_provisioning:
+                logger.info(
+                    '{0} - Resetting the number of consecutive '
+                    'read checks. Reason: scale up event detected'.format(
+                        table_name))
+                num_consec_read_checks = 0
                 update_needed = True
-                updated_read_units = updated_provisioning
+                updated_read_units = calculated_provisioning
 
     elif consumed_read_units_percent <= reads_lower_threshold:
 
         if decrease_reads_unit == 'percent':
-            updated_provisioning = calculators.decrease_reads_in_percent(
-                updated_read_units,
+            calculated_provisioning = calculators.decrease_reads_in_percent(
+                current_read_units,
                 decrease_reads_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'min_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
         else:
-            updated_provisioning = calculators.decrease_reads_in_units(
-                updated_read_units,
+            calculated_provisioning = calculators.decrease_reads_in_units(
+                current_read_units,
                 decrease_reads_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'min_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
 
-        if updated_read_units != updated_provisioning:
-            update_needed = True
-            updated_read_units = updated_provisioning
+        if current_read_units != calculated_provisioning:
+            # We need to look at how many times the num_consec_read_checks
+            # integer has incremented and Compare to config file value
+            num_consec_read_checks = num_consec_read_checks + 1
+
+            if num_consec_read_checks >= num_read_checks_before_scale_down:
+                update_needed = True
+                updated_read_units = calculated_provisioning
 
     if max_provisioned_reads:
         if int(updated_read_units) > int(max_provisioned_reads):
@@ -254,10 +305,16 @@ def __ensure_provisioning_reads(table_name, table_key, gsi_name, gsi_key):
                 'Will not increase writes over gsi-max-provisioned-reads '
                 'limit ({0} writes)'.format(updated_read_units))
 
-    return update_needed, int(updated_read_units)
+    logger.info('{0} - Consecutive read checks {1}/{2}'.format(
+        table_name,
+        num_consec_read_checks,
+        num_read_checks_before_scale_down))
+
+    return update_needed, updated_read_units, num_consec_read_checks
 
 
-def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
+def __ensure_provisioning_writes(
+        table_name, table_key, gsi_name, gsi_key, num_consec_write_checks):
     """ Ensure that provisioning of writes is correct
 
     :type table_name: str
@@ -268,7 +325,10 @@ def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
     :param gsi_name: Name of the GSI
     :type gsi_key: str
     :param gsi_key: Configuration option key name
-    :returns: (bool, int) -- update_needed, updated_write_units
+    :type num_consec_write_checks: int
+    :param num_consec_write_checks: How many consecutive checks have we had
+    :returns: (bool, int, int)
+        update_needed, updated_write_units, num_consec_write_checks
     """
     if not get_gsi_option(table_key, gsi_key, 'enable_writes_autoscaling'):
         logger.info(
@@ -276,11 +336,11 @@ def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
             'Autoscaling of writes has been disabled'.format(
                 table_name, gsi_name))
         return False, dynamodb.get_provisioned_gsi_write_units(
-            table_name, gsi_name)
+            table_name, gsi_name), 0
 
     update_needed = False
     try:
-        updated_write_units = dynamodb.get_provisioned_gsi_write_units(
+        current_write_units = dynamodb.get_provisioned_gsi_write_units(
             table_name, gsi_name)
         consumed_write_units_percent = \
             gsi_stats.get_consumed_write_units_percent(table_name, gsi_name)
@@ -303,11 +363,34 @@ def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
             get_gsi_option(table_key, gsi_key, 'decrease_writes_with')
         max_provisioned_writes = \
             get_gsi_option(table_key, gsi_key, 'max_provisioned_writes')
+        num_write_checks_before_scale_down = \
+            get_gsi_option(
+                table_key, gsi_key, 'num_write_checks_before_scale_down')
+        num_write_checks_reset_percent = \
+            get_gsi_option(table_key, gsi_key, 'num_write_checks_reset_percent')
     except JSONResponseError:
         raise
     except BotoServerError:
         raise
 
+    # Set the updated units to the current write unit value
+    updated_write_units = current_write_units
+
+    # Reset write consecutive count if num_write_checks_reset_percent is reached
+    if num_write_checks_reset_percent:
+
+        if consumed_write_units_percent >= num_write_checks_reset_percent:
+
+            logger.info(
+                '{0} - Resetting the number of consecutive '
+                'write checks. Reason: Consumed percent {1} is '
+                'greater than reset percent: {2}'.format(
+                    table_name,
+                    consumed_write_units_percent,
+                    num_write_checks_reset_percent))
+
+            num_consec_write_checks = 0
+
     # Check if we should update write provisioning
     if (consumed_write_units_percent == 0 and not get_gsi_option(
             table_key, gsi_key, 'allow_scaling_down_writes_on_0_percent')):
@@ -320,72 +403,75 @@ def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
     elif consumed_write_units_percent >= writes_upper_threshold:
 
         if increase_writes_unit == 'percent':
-            updated_provisioning = calculators.increase_writes_in_percent(
-                updated_write_units,
+            calculated_provisioning = calculators.increase_writes_in_percent(
+                current_write_units,
                 increase_writes_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
         else:
-            updated_provisioning = calculators.increase_writes_in_units(
-                updated_write_units,
+            calculated_provisioning = calculators.increase_writes_in_units(
+                current_write_units,
                 increase_writes_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
 
-        if updated_write_units != updated_provisioning:
+        if current_write_units != calculated_provisioning:
+            logger.info(
+                '{0} - Resetting the number of consecutive '
+                'write checks. Reason: scale up event detected'.format(
+                    table_name))
+            num_consec_write_checks = 0
             update_needed = True
-            updated_write_units = updated_provisioning
+            updated_write_units = calculated_provisioning
 
     elif throttled_write_count > throttled_writes_upper_threshold:
 
         if throttled_writes_upper_threshold > 0:
             if increase_writes_unit == 'percent':
-                updated_provisioning = calculators.increase_writes_in_percent(
-                    updated_write_units,
-                    increase_writes_with,
-                    table_name,
-                    table_key,
-                    gsi_name,
-                    gsi_key)
+                calculated_provisioning = \
+                    calculators.increase_writes_in_percent(
+                        current_write_units,
+                        increase_writes_with,
+                        get_gsi_option(
+                            table_key, gsi_key, 'max_provisioned_reads'),
+                        '{0} - GSI: {1}'.format(table_name, gsi_name))
             else:
-                updated_provisioning = calculators.increase_writes_in_units(
-                    updated_write_units,
+                calculated_provisioning = calculators.increase_writes_in_units(
+                    current_write_units,
                     increase_writes_with,
-                    table_name,
-                    table_key,
-                    gsi_name,
-                    gsi_key)
+                    get_gsi_option(table_key, gsi_key, 'max_provisioned_reads'),
+                    '{0} - GSI: {1}'.format(table_name, gsi_name))
 
-            if updated_write_units != updated_provisioning:
+            if current_write_units != calculated_provisioning:
+                logger.info(
+                    '{0} - Resetting the number of consecutive '
+                    'write checks. Reason: scale up event detected'.format(
+                        table_name))
+                num_consec_write_checks = 0
                 update_needed = True
-                updated_write_units = updated_provisioning
+                updated_write_units = calculated_provisioning
 
     elif consumed_write_units_percent <= writes_lower_threshold:
 
         if decrease_writes_unit == 'percent':
-            updated_provisioning = calculators.decrease_writes_in_percent(
-                updated_write_units,
+            calculated_provisioning = calculators.decrease_writes_in_percent(
+                current_write_units,
                 decrease_writes_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'min_provisioned_writes'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
         else:
-            updated_provisioning = calculators.decrease_writes_in_units(
-                updated_write_units,
+            calculated_provisioning = calculators.decrease_writes_in_units(
+                current_write_units,
                 decrease_writes_with,
-                table_name,
-                table_key,
-                gsi_name,
-                gsi_key)
+                get_gsi_option(table_key, gsi_key, 'min_provisioned_reads'),
+                '{0} - GSI: {1}'.format(table_name, gsi_name))
 
-        if updated_write_units != updated_provisioning:
-            update_needed = True
-            updated_write_units = updated_provisioning
+        if current_write_units != calculated_provisioning:
+            num_consec_write_checks = num_consec_write_checks + 1
+
+            if num_consec_write_checks >= num_write_checks_before_scale_down:
+                update_needed = True
+                updated_write_units = calculated_provisioning
 
     if max_provisioned_writes:
         if int(updated_write_units) > int(max_provisioned_writes):
@@ -399,41 +485,12 @@ def __ensure_provisioning_writes(table_name, table_key, gsi_name, gsi_key):
                     gsi_name,
                     updated_write_units))
 
-    return update_needed, int(updated_write_units)
+    logger.info('{0} - Consecutive write checks {1}/{2}'.format(
+        table_name,
+        num_consec_write_checks,
+        num_write_checks_before_scale_down))
 
-
-def __is_maintenance_window(table_name, gsi_name, maintenance_windows):
-    """ Checks that the current time is within the maintenance window
-
-    :type table_name: str
-    :param table_name: Name of the DynamoDB table
-    :type gsi_name: str
-    :param gsi_name: Name of the GSI
-    :type maintenance_windows: str
-    :param maintenance_windows: Example: '00:00-01:00,10:00-11:00'
-    :returns: bool -- True if within maintenance window
-    """
-    # Example string '00:00-01:00,10:00-11:00'
-    maintenance_window_list = []
-    for window in maintenance_windows.split(','):
-        try:
-            start, end = window.split('-', 1)
-        except ValueError:
-            logger.error(
-                '{0} - GSI: {1} - '
-                'Malformatted maintenance window'.format(table_name, gsi_name))
-            return False
-
-        maintenance_window_list.append((start, end))
-
-    now = datetime.datetime.utcnow().strftime('%H%M')
-    for maintenance_window in maintenance_window_list:
-        start = ''.join(maintenance_window[0].split(':'))
-        end = ''.join(maintenance_window[1].split(':'))
-        if now >= start and now <= end:
-            return True
-
-    return False
+    return update_needed, updated_write_units, num_consec_write_checks
 
 
 def __update_throughput(
@@ -461,24 +518,6 @@ def __update_throughput(
     except JSONResponseError:
         raise
 
-    # Check that we are in the right time frame
-    if get_gsi_option(table_key, gsi_key, 'maintenance_windows'):
-        if (not __is_maintenance_window(table_name, gsi_name, get_gsi_option(
-                table_key, gsi_key, 'maintenance_windows'))):
-
-            logger.warning(
-                '{0} - GSI: {1} - '
-                'Current time is outside maintenance window'.format(
-                    table_name,
-                    gsi_name))
-            return
-        else:
-            logger.info(
-                '{0} - GSI: {1} - '
-                'Current time is within maintenance window'.format(
-                    table_name,
-                    gsi_name))
-
     # Check table status
     try:
         gsi_status = dynamodb.get_gsi_status(table_name, gsi_name)
@@ -509,18 +548,10 @@ def __update_throughput(
                 table_name, gsi_name))
             return
 
-    if not get_global_option('dry_run'):
-        dynamodb.update_gsi_provisioning(
-            table_name,
-            table_key,
-            gsi_name,
-            gsi_key,
-            int(read_units),
-            int(write_units))
-        logger.info(
-            '{0} - GSI: {1} - '
-            'Provisioning updated to {2} reads and {3} writes'.format(
-                table_name,
-                gsi_name,
-                read_units,
-                write_units))
+    dynamodb.update_gsi_provisioning(
+        table_name,
+        table_key,
+        gsi_name,
+        gsi_key,
+        int(read_units),
+        int(write_units))
